# braces [![NPM version](https://badge.fury.io/js/braces.svg)](http://badge.fury.io/js/braces)

> Fastest brace expansion for node.js, with more complete support of Bash 4.3 than minimatch.

 - Expands comma-separated values: `a/{b,c}/d` => `['a/b/d', 'a/c/d']`
 - Expands alphabetical or numerical ranges: `{1..3}` => `['1', '2', '3']`
 - Better Bash 4.3 coverage than [minimatch](https://github.com/isaacs/minimatch), braces passes [most of the unit tests](#bash-4-3) from the [Bash 4.3 specification][bash], with a few [intentional differences](#differences).


## Install with [npm](npmjs.org)

```bash
npm i braces --save
```


## Example usage

```js
var braces = require('braces');

braces('a/{x,y}/c{d}e')
//=> ['a/x/cde', 'a/y/cde']

braces('a/b/c/{x,y}')
//=> ['a/b/c/x', 'a/b/c/y']

braces('a/{x,{1..5},y}/c{d}e')
//=> ['a/x/cde', 'a/1/cde', 'a/y/cde', 'a/2/cde', 'a/3/cde', 'a/4/cde', 'a/5/cde']
```

### Pro tip!

> Use braces to generate test fixtures!

**Example**

```js
var braces = require('./');
var path = require('path');
var fs = require('fs');

braces('blah/{a..z}.js').forEach(function(fp) {
  if (!fs.existsSync(path.dirname(fp))) {
    fs.mkdirSync(path.dirname(fp));
  }
  fs.writeFileSync(fp, '');
});
```

See the [tests](./test/test.js) for more examples and use cases (also see the [bash spec tests](./test/bash-mm-adjusted.js));


### Range expansion

Uses [expand-range](https://github.com/jonschlinkert/expand-range) for range expansion.

```js
braces('a{1..3}b')
//=> ['a1b', 'a2b', 'a3b']

braces('a{5..8}b')
//=> ['a5b', 'a6b', 'a7b', 'a8b']

braces('a{00..05}b')
//=> ['a00b', 'a01b', 'a02b', 'a03b', 'a04b', 'a05b']

braces('a{01..03}b')
//=> ['a01b', 'a02b', 'a03b']

braces('a{000..005}b')
//=> ['a000b', 'a001b', 'a002b', 'a003b', 'a004b', 'a005b']

braces('a{a..e}b')
//=> ['aab', 'abb', 'acb', 'adb', 'aeb']

braces('a{A..E}b')
//=> ['aAb', 'aBb', 'aCb', 'aDb', 'aEb']
```

Pass a function as the last argument to customize range expansions:

```js
var range = braces('x{a..e}y', function (str, i) {
  return String.fromCharCode(str) + i;
});

console.log(range);
//=> ['xa0y', 'xb1y', 'xc2y', 'xd3y', 'xe4y']
```

See [expand-range](https://github.com/jonschlinkert/expand-range) for benchmarks, tests and the full list of range expansion features.


## Bash 4.3

> Better support for Bash 4.3 than minimatch

This project has comprehensive unit tests, including tests coverted from [Bash 4.3][bash]. Currently only 8 of 102 unit tests fail, and  

### Differences

**Main differences from Bash:**

 - In bash (and minimatch) the braces are retained when only one set exists. ex: `a{b}c` => `['a{b}c']`. Since this can cause problems with routes and config templates etc., single-set valid braces are removed. e.g `a{b}c` => `['abc']`
 - duplicates are removed by default. To keep dupes, pass `nodupes: false` on the options

**New features:**

 - es6 delimiters are now ignored (e.g. not expanded as brace patterns), ex: `a/${b}/{c,d}` => `['a/${b}/c', 'a/${b}/d']`


## Benchmarks

```bash
node benchmark
```


## Run tests

```bash
npm test
```

## Contributing
Pull requests and stars are always welcome. For bugs and feature requests, [please create an issue](https://github.com/jonschlinkert/braces/issues).

Please run benchmarks before and after any code changes to what the impact of the code changes are before submitting a PR. 

## Author

**Jon Schlinkert**
 
+ [github/jonschlinkert](https://github.com/jonschlinkert)
+ [twitter/jonschlinkert](http://twitter.com/jonschlinkert) 

## License
Copyright (c) 2014-2015 Jon Schlinkert  
Released under the MIT license

***

_This file was generated by [verb](https://github.com/assemble/verb) on January 16, 2015._

[bash]: www.gnu.org/software/bash/
